package starj.toolkits.metrics;

import java.util.*;

import starj.EventBox;
import starj.coffer.*;
import starj.dependencies.*;
import starj.events.*;
import starj.spec.Constants;
import starj.toolkits.services.*;
import starj.util.*;

public class SynchronizationMetrics extends AbstractMetricOperation {
    public static final String CATEGORY = "concurrency";
    private int current_contention_index;
    private InstructionStartEvent inst_event; // cached event needed to request
                                              // a list of metric spaces for
                                              // a given call site
    private Set monitors;
    
    public SynchronizationMetrics(String name, String description) {
        super(name, description);
    }

    public EventDependencySet eventDependencies() {
        EventDependencySet dep_set = new EventDependencySet();
        dep_set.add(new EventDependency(
                Event.MONITOR_CONTENDED_ENTER,
                new TotalMask(Constants.FIELD_RECORDED
                        | Constants.FIELD_OBJECT),
                true));
        dep_set.add(new EventDependency(
                Event.MONITOR_CONTENDED_ENTERED,
                new TotalMask(Constants.FIELD_RECORDED
                        | Constants.FIELD_OBJECT),
                true));
        dep_set.add(new EventDependency(
                Event.INSTRUCTION_START,
                new TotalMask(Constants.FIELD_RECORDED),
                true));

        return dep_set;
    }
                        
    public OperationSet operationDependencies() {
        OperationSet dep_set = super.operationDependencies();
        dep_set.add(InstructionResolver.v());

        return dep_set;
    }

    public void init() {
        this.current_contention_index = ObjectEntity.registerCounter();
        this.monitors = new HashSet();
        this.inst_event = new InstructionStartEvent();
    }

    public void apply(EventBox box, MetricRecord[] records) {
        Event event = box.getEvent();
        switch (event.getID()) {
            case Event.MONITOR_CONTENDED_ENTER: {
                    MonitorContendedEnterEvent e
                            = (MonitorContendedEnterEvent) event;
                    ObjectEntity obj_entity = IDResolver.v().getObjectEntity(
                            e.getObjectID());

                    if (obj_entity != null) {
                        // Update contention counter
                        obj_entity.stepCounter(this.current_contention_index);
                    }
                    
                    InstructionContext ic
                            = InstructionResolver.v().getCurrentContext(
                                    e.getEnvID());
                    if (ic == null) {
                        break;
                    }
                    
                    InstructionStartEvent inst_event = this.inst_event;
                    ExecutionContext context = ic.getExecutionContext();
                    Instruction inst = ic.getInstruction();
                    if (inst == null || inst.getOpcode() != Code.MONITORENTER) {
                        break;
                    }
                     
                    inst_event.setEnvID(context.getThread().getID());
                    inst_event.setMethodID(context.getMethod().getID());
                    inst_event.setOffset(inst.getOffset());
                    MetricRecord[] inst_records
                            = this.getMetricRecords(inst_event);
                    for(int i = 0; i < inst_records.length; i++) {
                        SynchronizationMetricsRecord r
                                = (SynchronizationMetricsRecord)
                                        records[i];
                        inst.stepCounter(r.contention_count_index);
                        r.contention_count += 1L;
                    }
                    this.monitors.add(inst);
                }
                break;
            case Event.MONITOR_CONTENDED_ENTERED: {
                    MonitorContendedEnteredEvent e
                            = (MonitorContendedEnteredEvent) event;
                    ObjectEntity obj_entity = IDResolver.v().getObjectEntity(
                            e.getObjectID());

                    if (obj_entity != null) {
                        // Update contention counter
                        obj_entity.decrCounter(this.current_contention_index);
                    }
                }
                break;
            case Event.INSTRUCTION_START: {
                    InstructionStartEvent e = (InstructionStartEvent) event;
                    Instruction inst = e.getInstruction();
                    if (inst != null) {
                        if (inst.getOpcode() == Code.MONITORENTER) {
                            this.monitors.add(inst);
                            for(int i = 0; i < records.length; i++) {
                                SynchronizationMetricsRecord r
                                        = (SynchronizationMetricsRecord)
                                                records[i];
                                r.monitorenter_count += 1L;
                                inst.stepCounter(r.entry_count_index);
                            }
                        }
                    }
                }
                break;
            default:
                break;
        }
    }

    public void accept(MetricVisitor visitor, MetricRecord record) {
        SynchronizationMetricsRecord r = (SynchronizationMetricsRecord) record;

        long inst_count = InstructionCountManager.v().getInstructionCount(
                record);
        visitor.visit(new ValueMetric(CATEGORY, "lockDensity",
                    new DensityMetricValue(r.monitorenter_count,
                            inst_count)));
        
        visitor.visit(new ValueMetric(CATEGORY, "contendedLockDensity",
                new DensityMetricValue(r.contention_count, inst_count)));
        
        float threshold = 0.90f;
        
        WeightFunction monitor_entry_function = new CounterWeightFunction(
                r.entry_count_index);
        Collection touched_monitors = CollectionUtils.filter(this.monitors,
                new WeightFilter(monitor_entry_function));
        visitor.visit(new PercentileMetric(CATEGORY, "lock",
                new PercentageMetricValue(PercentileManager.getPercentile(
                        touched_monitors, monitor_entry_function,
                        threshold)), threshold));
        touched_monitors = null;
        
        WeightFunction contention_function
                = new CounterWeightFunction(r.contention_count_index);
        Collection contended_monitors = CollectionUtils.filter(this.monitors,
                new WeightFilter(contention_function));
        visitor.visit(new PercentileMetric(CATEGORY, "contendedLock",
                new PercentageMetricValue(PercentileManager.getPercentile(
                        contended_monitors, contention_function,
                        threshold)), threshold));
        contended_monitors = null;
    }
    
    public MetricRecord newRecord() {
        return new SynchronizationMetricsRecord();
    }
    
    private class SynchronizationMetricsRecord implements MetricRecord {
        public long monitorenter_count;
        public long contention_count;
        public int entry_count_index;
        public int contention_count_index;
        
        public SynchronizationMetricsRecord() {
            this.entry_count_index = Instruction.registerCounter();
            this.contention_count_index = Instruction.registerCounter();
        }
    }
}
